# -*- coding:utf-8 -*-

"""
Created on 25-05-2012

Author: Edwin Dalmaijer

This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.

"""
from libopensesame.py3compat import *
import pygame
from .basejoystick import BaseJoystick


class Legacy(BaseJoystick):

    def __init__(self, experiment, device=0, joybuttonlist=None, timeout=None):
        """See _libjoystick.basejoystick"""
        self.js = pygame.joystick.Joystick(device)
        self.js.init()
        self.experiment = experiment
        self.set_joybuttonlist(joybuttonlist)
        self.set_timeout(timeout)
        pygame.event.set_blocked(pygame.JOYAXISMOTION)
        pygame.event.set_blocked(pygame.JOYHATMOTION)
        pygame.event.set_blocked(pygame.JOYBALLMOTION)

    def get_joybutton(self, joybuttonlist=None, timeout=None):
        """See _libjoystick.basejoystick"""
        if joybuttonlist is None or joybuttonlist == []:
            joybuttonlist = self._joybuttonlist
        if timeout is None:
            timeout = self.timeout
        start_time = pygame.time.get_ticks()
        time = start_time
        while timeout is None or time - start_time <= timeout:
            time = pygame.time.get_ticks()
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        self.experiment.pause()
                if event.type == pygame.JOYBUTTONDOWN:
                    if (
                            joybuttonlist is None or
                            event.button + 1 in joybuttonlist
                    ):
                        bpress = event.button + 1
                        return bpress, time
        return None, time

    def get_joyaxes(self, timeout=None):
        """See _libjoystick.basejoystick"""
        pygame.event.set_allowed(pygame.JOYAXISMOTION)
        if timeout is None:
            timeout = self.timeout
        pos = []
        start_time = pygame.time.get_ticks()
        time = start_time
        while timeout is None or time - start_time < timeout:
            time = pygame.time.get_ticks()
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        self.experiment.pause()
                if event.type == pygame.JOYAXISMOTION:
                    for axis in range(self.js.get_numaxes()):
                        pos.append(self.js.get_axis(axis))
                    pygame.event.set_blocked(pygame.JOYAXISMOTION)
                    return pos, time
        pygame.event.set_blocked(pygame.JOYAXISMOTION)
        return None, time

    def get_joyballs(self, timeout=None):
        """See _libjoystick.basejoystick"""
        pygame.event.set_allowed(pygame.JOYBALLMOTION)
        if timeout is None:
            timeout = self.timeout
        ballpos = []
        start_time = pygame.time.get_ticks()
        time = start_time
        while timeout is None or time - start_time < timeout:
            time = pygame.time.get_ticks()
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        self.experiment.pause()
                if event.type == pygame.JOYBALLMOTION:
                    for ball in range(self.js.get_numballs()):
                        ballpos.append(self.js.get_ball(ball))
                    pygame.event.set_blocked(pygame.JOYBALLMOTION)
                    return ballpos, time
        pygame.event.set_blocked(pygame.JOYBALLMOTION)
        return None, time

    def get_joyhats(self, timeout=None):
        """See _libjoystick.basejoystick"""
        pygame.event.set_allowed(pygame.JOYHATMOTION)
        if timeout is None:
            timeout = self.timeout
        hatpos = []
        start_time = pygame.time.get_ticks()
        time = start_time
        while timeout is None or time - start_time < timeout:
            time = pygame.time.get_ticks()
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        self.experiment.pause()
                if event.type == pygame.JOYHATMOTION:
                    for hat in range(self.js.get_numhats()):
                        hatpos.append(self.js.get_hat(hat))
                    pygame.event.set_blocked(pygame.JOYHATMOTION)
                    return hatpos, time
        pygame.event.set_blocked(pygame.JOYHATMOTION)
        return None, time

    def get_joyinput(self, joybuttonlist=None, timeout=None):
        """See _libjoystick.basejoystick"""
        pygame.event.set_allowed(pygame.JOYHATMOTION)
        pygame.event.set_allowed(pygame.JOYAXISMOTION)
        pygame.event.set_allowed(pygame.JOYBALLMOTION)
        if joybuttonlist is None or joybuttonlist == []:
            joybuttonlist = self._joybuttonlist
        if timeout is None:
            timeout = self.timeout
        pos = []
        ballpos = []
        hatpos = []
        eventtype = None
        start_time = pygame.time.get_ticks()
        time = start_time
        while timeout is None or time - start_time <= timeout:
            time = pygame.time.get_ticks()
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        self.experiment.pause()
                if event.type == pygame.JOYBUTTONDOWN:
                    if (
                            joybuttonlist is None or
                            event.button + 1 in joybuttonlist
                    ):
                        eventtype = u'joybuttonpress'
                        bpress = event.button + 1
                        pygame.event.set_blocked(pygame.JOYHATMOTION)
                        pygame.event.set_blocked(pygame.JOYAXISMOTION)
                        pygame.event.set_blocked(pygame.JOYBALLMOTION)
                        return eventtype, bpress, time
                if event.type == pygame.JOYAXISMOTION:
                    eventtype = u'joyaxismotion'
                    for axis in range(self.js.get_numaxes()):
                        pos.append(self.js.get_axis(axis))
                    pygame.event.set_blocked(pygame.JOYHATMOTION)
                    pygame.event.set_blocked(pygame.JOYAXISMOTION)
                    pygame.event.set_blocked(pygame.JOYBALLMOTION)
                    return eventtype, pos, time
                if event.type == pygame.JOYBALLMOTION:
                    eventtype = u'joyballmotion'
                    for ball in range(self.js.get_numballs()):
                        ballpos.append(self.js.get_ball(ball))
                    pygame.event.set_blocked(pygame.JOYHATMOTION)
                    pygame.event.set_blocked(pygame.JOYAXISMOTION)
                    pygame.event.set_blocked(pygame.JOYBALLMOTION)
                    return eventtype, ballpos, time
                if event.type == pygame.JOYHATMOTION:
                    eventtype = u'joyhatmotion'
                    for hat in range(self.js.get_numhats()):
                        hatpos.append(self.js.get_hat(hat))
                    pygame.event.set_blocked(pygame.JOYHATMOTION)
                    pygame.event.set_blocked(pygame.JOYAXISMOTION)
                    pygame.event.set_blocked(pygame.JOYBALLMOTION)
                    return eventtype, hatpos, time
        pygame.event.set_blocked(pygame.JOYHATMOTION)
        pygame.event.set_blocked(pygame.JOYAXISMOTION)
        pygame.event.set_blocked(pygame.JOYBALLMOTION)
        return eventtype, None, time

    def input_options(self):
        """See _libjoystick.basejoystick"""
        ninputs = [
            self.js.get_numbuttons(),
            self.js.get_numaxes(),
            self.js.get_numballs(),
            self.js.get_numhats()
        ]
        return ninputs

    def flush(self):
        """See _libjoystick.basejoystick"""
        joyinput = False
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
                self.experiment.pause()
            if event.type in (
                    pygame.JOYBUTTONDOWN,
                    pygame.JOYAXISMOTION,
                    pygame.JOYBALLMOTION,
                    pygame.JOYHATMOTION
            ):
                joyinput = True
        return joyinput



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._ellipse.ellipse import Ellipse
from openexp._canvas._element.psycho import PsychoElement
from psychopy import visual
from PIL import Image, ImageDraw
import numpy as np


class Psycho(PsychoElement, Ellipse):

    def prepare(self):

        p = int(self.penwidth)
        w = int(self.w)
        h = int(self.h)
        x = int(self.x)
        y = int(self.y)
        if self.fill:
            # A filled ellipse ignores the penwidth and fills up exactly the
            # specified size
            im = Image.new('L', (w, h), 0)
            dr = ImageDraw.Draw(im)
            dr.ellipse((0, 0, w-1, h-1), fill=1)
        elif p == 1:
            # An unfilled ellipses with a pendwidth of 1 also fills up exactly
            # the specified size
            im = Image.new('L', (w, h), 0)
            dr = ImageDraw.Draw(im)
            dr.ellipse((0, 0, w-1, h-1), outline=1)
        else:
            # An unfilled ellipse with a positive larger than 1 fills up
            # slightly more than the specified size, because half the penwidth
            # extends beyond the size on each side.
            im = Image.new('L', (w+p, h+p), 0)
            dr = ImageDraw.Draw(im)
            dr.ellipse((0, 0, w+p-1, h+p-1), fill=1)
            dr.ellipse((p, p, w-1, h-1), fill=0)
        # Transform the image into a mask, and remap the colors to PsychoPy
        # format (-1 = transparent, 1 = opaque)
        mask = np.array(im, dtype=int)
        mask[mask == 0] = -1
        self._stim = visual.GratingStim(win=self.win, mask=mask,
                                        pos=self.to_xy(x+w/2, y+h/2), size=im.size,
                                        color=self.color.backend_color, tex=None)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._polygon.polygon import Polygon
from openexp._canvas._element.psycho import PsychoElement
from psychopy import visual


class Psycho(PsychoElement, Polygon):

    def prepare(self):

        self._stim = visual.ShapeStim(
            self.win,
            lineWidth=self.penwidth,
            vertices=[self.to_xy(x, y) for x, y in self.vertices],
            lineColor=self.color.backend_color,
            closeShape=True,
            fillColor=self.color.backend_color if self.fill else None,
            interpolate=False
        )



"""Shows the get-started tab and opens an experiment on startup, if one was
passed on the command line
"""
icon = "document-new"
label = "New\u2026"
priority = -1
shortcut = "Ctrl+N"
menu = {
    "index": 0,
    "separator_after": False,
    "separator_before": False,
    "submenu": "File"
}
toolbar = {
    "index": 0,
    "separator_after": False,
    "separator_before": False
}
tooltip = "Start a new experiment"



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue, InvalidFormScript
from libopensesame.item import Item
from libopensesame import widgets
from libqtopensesame.items.qtautoplugin import QtAutoPlugin


class FormMultipleChoice(Item):

    lazy_init = False

    def reset(self):
        self.var.options = u'Yes\nNo\nMaybe'
        self.var.question = u'Your question'
        self.var.form_title = u'Form title'
        self.var.form_var = u'response'
        self.var.advance_immediately = u'yes'
        self.var.allow_multiple = u'yes'
        self.var.button_text = u'Ok'
        self.var.timeout = u'infinite'
        self.var.spacing = 10
        self.var.margins = u'50;50;50;50'
        self.var._theme = u'gray'

    def run(self):
        # Parse the option list
        option_list = self.var.options.split(u'\n')  # split by return
        # Filter out empty options
        option_list = list(filter(lambda option: option != u'', option_list))
        if not option_list:
            raise InvalidFormScript(
                'You must specify at least one response option')
        # Determine whether a button is shown and determine the number of rows
        rows = len(option_list) + 2
        if (
                self.var.advance_immediately == u'no' or
                self.var.allow_multiple == u'yes'
        ):
            show_button = True
            click_accepts = False
            rows += 1
        else:
            show_button = False
            click_accepts = True
        # Determine the group for the checkboxes
        if self.var.allow_multiple == u'no':
            group = u'response_group'
        else:
            group = None
        # The variable in which the response is stored
        var = self.var.form_var
        # Build the form
        try:
            margins = [
                float(i) for i in safe_decode(self.var.margins).split(u';')
            ]
        except (ValueError, TypeError):
            raise InvalidValue(
                'margins should be numeric values separated by a semi-colon')
        if self.var.timeout == u'infinite':
            timeout = None
        else:
            timeout = self.var.timeout
        form = widgets.form(
            self.experiment,
            cols=1,
            rows=rows,
            spacing=self.var.spacing,
            margins=margins,
            theme=self.var._theme,
            item=self,
            timeout=timeout,
            clicks=self.var.form_clicks == u'yes'
        )
        form.set_widget(widgets.label(form, self.var.form_title), (0, 0))
        form.set_widget(widgets.label(form, self.var.question), (0, 1))
        i = 2
        for option in option_list:
            form.set_widget(
                widgets.checkbox(
                    form,
                    option,
                    group=group,
                    click_accepts=click_accepts,
                    var=var
                ),
                (0, i)
            )
            i += 1
        if show_button:
            form.set_widget(widgets.button(form, self.var.button_text), (0, i))
        form._exec()

    def var_info(self):
        return (super().var_info() + 
                [(self.var.form_var, u'[Depends on response]')])


class QtFormMultipleChoice(FormMultipleChoice, QtAutoPlugin):

    def __init__(self, name, experiment, string=None):
        FormMultipleChoice.__init__(self, name, experiment, string)
        QtAutoPlugin.__init__(self, __file__)
        self.custom_interactions()

    def apply_edit_changes(self):
        if not super().apply_edit_changes() or self.lock:
            return False
        self.custom_interactions()
        return True

    def custom_interactions(self):
        """
        The advance_immediately option is not applicable if multiple items can
        be selected.
        """
        self.checkbox_advance_immediately.setEnabled(
            self.var.get(u'allow_multiple') == u'no')
        self.line_edit_button_text.setEnabled(
            self.var.get(u'allow_multiple') == u'yes' or
            self.var.get(u'advance_immediately') == u'no')



"""Collects input from a joystick or gamepad"""

authors = ["Edwin Dalmaijer", "Sebastiaan Mathot"]
category = "Response collection"
icon = "os-joystick"
controls = [
    {
        "type": "combobox",
        "var": "_dummy",
        "label": "Dummy mode (use keyboard instead of joystick)",
        "options": [
            "no",
            "yes"
        ],
        "tooltip": "Enable dummy mode to test the experiment using a keyboard"
    },
    {
        "type": "spinbox",
        "var": "_device",
        "label": "Device nr.",
        "min_val": 0,
        "max_val": 100,
        "tooltip": "Identifies the joystick, in case there are multiple joysticks"
    },
    {
        "type": "line_edit",
        "var": "correct_response",
        "label": "Correct response",
        "tooltip": "Expecting a comma-separated list of numbers between 1 and the number of joybuttons"
    },
    {
        "type": "line_edit",
        "var": "allowed_responses",
        "label": "Allowed responses",
        "tooltip": "Expecting a comma-separated list of numbers between 1 and the number of joybuttons"
    },
    {
        "type": "line_edit",
        "var": "timeout",
        "label": "Timeout",
        "tooltip": "Expecting a value in milliseconds of 'infinite'"
    }
]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import re
from libopensesame.oslogging import oslogger
from pyqode.core import api
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'pyqode_manager', category=u'extension')

THRESHOLD = 500


class AutodetectIndentationMode(api.Mode):

    """Automatically sets indentation based on editor contents."""
    def on_install(self, editor):

        super(AutodetectIndentationMode, self).on_install(editor)
        self.editor.new_text_set.connect(self._autodetect_indentation)
        self._autodetect_indentation()

    def _autodetect_indentation(self):

        if cfg.pyqode_indentation == 'spaces':
            self._set_space_indentation(cfg.pyqode_tab_length)
            return
        if cfg.pyqode_indentation == 'tabs':
            self._set_tab_indentation()
            return
        code = self.editor.toPlainText().replace(u'\u2029', u'\n')
        match_tabs = re.findall(u'((?<=\n)\t+)|(\A\t+)', code)
        tab_length = self._guess_indentation_width(code)
        if tab_length == 0:
            match_spaces = []
        else:
            match_spaces = re.findall(
                u'((?<=\n){indent}+)|(\A{indent}+)'.format(
                    indent=u' ' * tab_length
                ),
                code
            )
        if len(match_tabs) > len(match_spaces):
            oslogger.debug(u'detected tab-based indentation')
            self._set_tab_indentation()
        else:
            oslogger.debug(
                u'detected space-based indentation (width={})'.format(
                    tab_length
                )
            )
            self._set_space_indentation(tab_length)

    def _set_tab_indentation(self):

        self.editor.use_spaces_instead_of_tabs = False

    def _set_space_indentation(self, tab_length):

        self.editor.use_spaces_instead_of_tabs = True
        self.editor.tab_length = tab_length

    def _guess_indentation_width(self, code):

        i2 = i3 = i4 = 0
        for line in code.splitlines():
            n_indent = len(line) - len(line.lstrip(u' '))
            if not n_indent:
                continue
            if not n_indent % 4:
                i4 += 1
                i3 -= 1
            elif not n_indent % 2:
                i2 += 1
                i4 -= 1
                i3 -= 1
            elif not n_indent % 3:
                i3 += 1
                i2 -= 1
                i4 -= 1
            if i4 - i2 > THRESHOLD and i4 - i3 > THRESHOLD:
                return 4
            if i3 - i2 > THRESHOLD and i3 - i4 > THRESHOLD:
                return 3
            if i2 - i3 > THRESHOLD and i2 - i4 > THRESHOLD:
                return 2
        if i4 > max(i2, i3):
            return 4
        if i3 > max(i2, i4):
            return 3
        if i2 > max(i4, i3):
            return 2
        return cfg.pyqode_tab_length



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent


class QtItemSplitter(BaseSubcomponent, QtWidgets.QSplitter):

    r"""Implements a splitter for the edit and script view in an item tab. This
    custom is mostly necessary, because the default QSplitter.sizeHint() is too
    large.
    """
    def __init__(self, item):
        r"""Constructor."""
        super().__init__(QtCore.Qt.Vertical, item.main_window)
        self.item = item
        self.setup(item.main_window)
        self.addWidget(self.item._edit_widget)
        self.addWidget(self.item._script_frame)
        self.splitterMoved.connect(self._on_splitter_moved)

    def _on_splitter_moved(self, pos, index):

        self.splitterMoved.disconnect()
        if self.item._script_widget is None:
            self.item.init_script_widget()

    def minimumSizeHint(self):
        """
        returns:
                type:	QSize
        """
        return QtCore.QSize(100, 100)

    def sizeHint(self):
        """
        returns:
                type:	QSize
        """
        return QtCore.QSize(100, 100)


# Alias for backwards compatibility
qtitem_splitter = QtItemSplitter



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import ItemDoesNotExist, \
    ConditionalExpressionError
from libopensesame.item import Item
import gc
from openexp.keyboard import Keyboard


class Sequence(Item):

    description = 'Runs a number of items in sequence'

    def reset(self):
        self.items = []
        self.var.flush_keyboard = 'yes'

    def run(self):
        self.set_item_onset()
        # Optionally flush the responses to catch escape presses
        if self._keyboard is not None:
            self._keyboard.flush()
        for _item, cond in self._items:
            self.python_workspace[u'self'] = self
            try:
                run = self.python_workspace._eval(cond)
            except Exception as e:
                raise ConditionalExpressionError(
                    'Error while evaluating run-if expression')
            if run:
                self.experiment.items.run(_item)
        if not gc.isenabled():
            gc.collect()

    def set_validator(self):
        self.validator = sequence

    def from_string(self, string):
        self.var.clear()
        self.comments = []
        self.reset()
        if string is None:
            return
        for s in string.split(u'\n'):
            self.parse_variable(s)
            cmd, arglist, kwdict = self.syntax.parse_cmd(s)
            if cmd != 'run' or not len(arglist):
                continue
            _item = arglist[0]
            if len(arglist) == 1:
                cond = 'True'
            else:
                cond = arglist[1]
            enabled = len(arglist) < 3 or arglist[2].lower() != 'disabled'
            self.items.append((_item, cond, enabled))

    def prepare(self):
        super().prepare()
        if self.var.flush_keyboard == 'yes':
            # Create a keyboard to flush responses at the start of the run
            # phase
            self._keyboard = Keyboard(self.experiment)
        else:
            self._keyboard = None
        self._items = []
        for _item, cond, enabled in self.items:
            if not enabled:
                continue
            if _item not in self.experiment.items:
                raise ItemDoesNotExist(_item)
            self.experiment.items.prepare(_item)
            self._items.append((_item, self.syntax.compile_cond(cond)))

    def to_string(self):
        s = super().to_string(self.item_type)
        for _item, cond, enabled in self.items:
            arglist = [_item, cond]
            if not enabled:
                arglist.append('disabled')
            s += '\t' + self.syntax.create_cmd('run', arglist) + '\n'
        return s


# Alias for backwards compatibility
sequence = Sequence



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame import misc
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame._input.popup_menu import PopupMenu
from libqtopensesame._input.confirmation import Confirmation
from qtpy import QtCore, QtWidgets
import os
import os.path
import shutil
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'pool_widget', category=u'core')


class PoolWidget(BaseWidget):
    """The file-pool widget.
    
    Parameters
    ----------
    main_window
        The main-window object.
    """
    def __init__(self, main_window):
        self.max_len = 5
        super().__init__(main_window, ui=u'widgets.pool_widget')
        self.ui.button_pool_add.clicked.connect(self.select_and_add)
        self.ui.button_refresh.clicked.connect(self.refresh)
        self.ui.button_help_pool.clicked.connect(self.help)
        self.ui.button_browse_pool.clicked.connect(self.browse)
        self.ui.edit_pool_filter.textChanged.connect(self.refresh)
        self.ui.combobox_view.currentIndexChanged.connect(self.set_view)
        self.ui.list_pool.itemActivated.connect(self.activate_file)
        self.ui.list_pool.itemChanged.connect(self.rename_file)
        self.main_window.theme.apply_theme(self)
        self.ui.label_size_warning.setVisible(False)

    def help(self):
        """Opens the help tab."""
        self.tabwidget.open_osdoc(u'manual/interface')

    def set_view(self):
        """Sets the viewmode (list/ icons) based on the gui control."""
        if self.ui.combobox_view.currentIndex() == 0:
            self.ui.list_pool.setViewMode(QtWidgets.QListView.ListMode)
        else:
            self.ui.list_pool.setViewMode(QtWidgets.QListView.IconMode)

    def browse(self):
        """Opens the pool folder in the file manager in an OS specific way."""
        misc.open_url(self.experiment.pool.folder())

    def add(self, files, rename=False):
        """Adds a list of files to the pool.

        Parameters
        ----------
        files : list
            A list of paths.
        """
        if not files:
            return
        n_replace = sum(
            self.pool.in_folder(os.path.basename(path))
            for path in files
        )
        if not rename and n_replace:
            if n_replace == 1:
                msg = _(
                    u"The file pool already contains a file with the same name. Do you want to overwrite it?")
            else:
                msg = _(
                    u"The file pool already contains files with the same names. Do you want to overwrite {} files?")
            c = Confirmation(self.main_window, msg.format(n_replace))
            if not c.show():
                return
        for path in files:
            basename = os.path.basename(path)
            if self.pool.in_folder(basename) and rename:
                while self.pool.in_folder(basename):
                    basename = u'_' + basename
            try:
                self.pool.add(path, new_name=basename)
            except (IOError, shutil.Error) as e:
                self.notify(_(u'Failed to copy %s to file pool') % path)
                self.console.write(safe_decode(e, errors=u'ignore'))
        self.main_window.set_unsaved()
        self.refresh()
        self.select(basename)

    def select_and_add(self, dummy=None):
        """Adds one or more files to the pool."""
        path_list = QtWidgets.QFileDialog.getOpenFileNames(
            self.main_window.ui.centralwidget, _(u"Add files to pool"),
            directory=cfg.default_pool_folder)
        if not path_list:
            return
        # Qt5 puts the first element in another list, check for that here
        if isinstance(path_list[0], list):
            path_list = path_list[0]
        if not path_list:
            return
        cfg.default_pool_folder = os.path.dirname(path_list[0])
        self.add(path_list)

    def select(self, fname):
        """Selects a specific file in the file pool.

        Parameters
        ----------
        fname : str, unicode
            The file to be selected.
        """
        for i in range(self.ui.list_pool.count()):
            item = self.ui.list_pool.item(i)
            if item.text() == fname:
                self.ui.list_pool.setCurrentItem(item)

    def refresh(self):
        """Refreshes the contents of the pool widget."""
        try:
            path_iterator = iter(self.pool)
        except Exception as e:
            self.notify(_(u'Failed to refresh file pool'))
            self.console.write(safe_decode(e, errors=u'ignore'))
            return
        filt = self.ui.edit_pool_filter.text().lower()
        self.ui.list_pool.clear()
        for path in path_iterator:
            oslogger.debug(path)
            fname = os.path.basename(path)
            if filt in fname.lower():
                icon = self.theme.qfileicon(self.pool[path])
                item = QtWidgets.QListWidgetItem(icon, fname)
                item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable)
                item.icon = icon
                item.path = path
                item.setToolTip(path)
                self.ui.list_pool.addItem(item)
        try:
            size = self.pool.size()
        except:
            size = -1
        oslogger.debug(u'pool is %d bytes' % size)
        if size > cfg.file_pool_size_warning:
            self.ui.label_size_warning.setText(_('Your file pool is larger '
                                                 'than usual (%d MB). This increases loading and saving time. '
                                                 'Consider moving files from the file pool to the experiment '
                                                 'folder.') % (size / 1048576))
            self.ui.label_size_warning.setVisible(True)
        else:
            self.ui.label_size_warning.setVisible(False)

    def open_file(self, path):
        """Opens a file in a platform specific way.

        Parameters
        ----------
        path : unicode, str
            The file to be opened.
        """
        misc.open_url(self.pool[path])

    def activate_file(self, item):
        """Is called when a file is double-clicked or otherwise activated and
        opens the file.

        Parameters
        ----------
        item : QListWidgetItem
        """
        self.open_file(item.path)

    def contextMenuEvent(self, event):
        """Presents a context menu.

        Parameters
        ----------
        event : QMouseClickEvent
        """
        item = self.ui.list_pool.itemAt(self.ui.list_pool.mapFromGlobal(
            event.globalPos()))
        if item is None:
            return
        path = item.path
        pm = PopupMenu(self, [
            (0, _(u'Open'), item.icon),
            (1, _(u'Remove from pool'), u'delete'),
            (2, _(u'Rename'), u'rename'),
        ])
        action = pm.show()
        if action == 0:
            self.open_file(path)
        elif action == 1:
            self.delete_selected_files()
        elif action == 2:
            self.ui.list_pool.editItem(item)

    def delete_selected_files(self):
        """Deletes all selected files from the pool. Asks for confirmation\
        first.
        """
        # Prepare the confirmation dialog, which contains a limited nr of
        # filenames
        l = []
        suffix = u''
        for item in self.ui.list_pool.selectedItems()[:self.max_len]:
            l.append(self.pool[item.text()])
        if len(self.ui.list_pool.selectedItems()) > self.max_len:
            suffix = _('And %d more file(s)') % \
                (len(self.ui.list_pool.selectedItems())-self.max_len)
        msg = _(u"<p>Are you sure you want to remove the following files?</p>"
                u"<p>- %s</p> <p>%s</p>") % (u"<br /> - ".join(l), suffix)
        c = Confirmation(self.main_window, msg)
        if not c.show():
            return
        # Create a list of files to be removed
        dL = []
        for item in self.ui.list_pool.selectedItems():
            dL.append(item.path)
        # Remove the files
        try:
            for f in dL:
                del self.pool[f]
            oslogger.debug(u"removed '%s'" % f)
        except:
            oslogger.error(u"failed to remove '%s'" % f)
        self.refresh()
        self.main_window.set_unsaved()

    def rename_file(self, item):
        """Starts a rename action for an item.

        Parameters
        ----------
        item : QListWidgetItem
        """
        old_name = item.path
        new_name = item.text().strip()
        if new_name in self.pool:
            self.notify(
                _(u"There already is a file named '%s' in the file pool")
                % new_name)
            new_name = old_name
        elif new_name in (old_name, u''):
            new_name = old_name
        else:
            try:
                self.pool.rename(old_name, new_name)
            except:
                self.notify(_(u'Failed to rename "%s" to "%s".')
                            % (old_name, new_name))
            else:
                self.main_window.set_unsaved()
        self.refresh()
        self.select(new_name)

    def dragEnterEvent(self, event):
        """Accepts an incoming drag that precedes a drop.

        Parameters
        ----------
        event : QDragEnterEvent
        """
        event.acceptProposedAction()

    def dropEvent(self, event):
        """Accepts an incoming drop.

        Parameters
        ----------
        event : QDropEnterEvent
        """
        files = []
        for url in event.mimeData().urls():
            files.append(url.toLocalFile())
        self.add(files)
        event.acceptProposedAction()

    def focusInEvent(self, e):
        """Focus the filter edit when the widget receives focus.

        Parameters
        ----------
        e : QFocusEvent
        """
        self.ui.edit_pool_filter.setFocus()


def select_from_pool(main_window, parent=None):
    """A static function that presents the select from pool dialog.

    Parameters
    ----------
    main_window
        The GUI main window
    parent, optional
        The parent QWidget or None to use main window's central widget.
    """
    if parent is None:
        parent = main_window.ui.centralwidget
    d = QtWidgets.QDialog(parent)
    widget = PoolWidget(main_window)
    widget.refresh()
    bbox = QtWidgets.QDialogButtonBox(d)
    bbox.addButton(_(u"Cancel"), QtWidgets.QDialogButtonBox.RejectRole)
    bbox.addButton(_(u"Select"), QtWidgets.QDialogButtonBox.AcceptRole)
    bbox.accepted.connect(d.accept)
    bbox.rejected.connect(d.reject)
    vbox = QtWidgets.QVBoxLayout()
    vbox.addWidget(widget)
    vbox.addWidget(bbox)
    d.setLayout(vbox)
    d.setWindowTitle(_(u"Select file from pool"))
    res = d.exec_()
    main_window.ui.pool_widget.refresh()
    if res == QtWidgets.QDialog.Rejected:
        return u""
    selected = widget.ui.list_pool.currentItem()
    if selected is None:
        return u""
    return selected.text()


# Alias for backwards compatibility
pool_widget = PoolWidget



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.

Translation logic:

- Context naming:
    - main code: core_[name]
    - items: item_[name]
    - plugins: plugin_[name]
    - extensions: extension_[name]
    - UI files: ui_[name]
- Literals are translated where they are defined
- Wildcards in literals should be part of the translation string
- Translatables should not be broken across lines
- For plugns and extensions, JSON info is extracted,
  and translated when automatic controls are defined
- UI files are extracted, and translated fully automatically
"""
from libopensesame.py3compat import *
from qtpy.QtCore import QCoreApplication


def translation_context(name, category=u'core'):
    r"""A factory that generates a translation method with a fixed context.

    Parameters
    ----------
    name : str
        The name to use for the context.
    name : str, optional
        The category to use for the context.

    Returns
    -------
    FunctionType
        A translation function.
    """
    _context = u'%s_%s' % (category, name)
    return lambda s, context=None: QCoreApplication.translate(_context, s)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from qtpy import QtWidgets


class BaseWidget(QtWidgets.QWidget, BaseSubcomponent):
    r"""A base class for widgets.

    Parameters
    -----------
    main_window : QtOpenSesame
    ui : str, optional
        An id for a user-interface file, for example 'dialogs.quick_switcher'.
    *arglist:
        passed onto parent constructors.
    *kwdict:
        passed onto parent constructors.
    """
    
    def __init__(self, parent, ui=None, *arglist, **kwdict):
        super().__init__(parent, *arglist, **kwdict)
        self.setup(parent, ui=ui)


# Alias for backwards compatibility
base_widget = BaseWidget



"""Integrates the menu into the toolbar."""

checkable = True
icon = "applications-system"
label = "Integrate menu and toolbar"
menu = {
    "index": -1,
    "separator_after": False,
    "separator_before": True,
    "submenu": "View"
}
priority = -999999
settings = {
    "toolbar_menu_active": False
}
tooltip = "Integrates the menu into the toolbar"



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from libqtopensesame.dialogs.base_dialog import BaseDialog
from datamatrix import DataMatrix, operations
from qdatamatrix import QDataMatrix
import pickle
from qtpy import QtWidgets


class LoopWizard(BaseDialog):

    r"""The loop-wizard dialog"""
    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window
            The main window object.
        """
        super().__init__(main_window, ui=u'dialogs.loop_wizard_dialog')
        try:
            self._dm = pickle.loads(cfg.loop_wizard)
            assert(isinstance(self._dm, DataMatrix))
        except Exception as e:
            oslogger.warning(f'failed to restore loop table: {e}')
            self._dm = DataMatrix(length=0)
        self._qdm = QDataMatrix(self._dm)
        self.ui.hbox_container.addWidget(self._qdm)
        self.ui.table_example.hide()

    def exec_(self):
        r"""Executes the dialog.

        Returns
        -------
        The dialog return status.
        """
        retval = super().exec_()
        if retval == QtWidgets.QDialog.Rejected:
            return None
        self._dm = self._qdm.dm
        cfg.loop_wizard = pickle.dumps(self._dm)
        return operations.fullfactorial(self._dm)


# Alias for backwards compatibility
loop_wizard = LoopWizard



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp.backend import Backend, configurable
from libopensesame.exceptions import InvalidValue


class Sampler(Backend):
    r"""The `Sampler` class provides functionality to play sound samples. You 
    generally create a `Sampler` object with the `Sampler()` factory function, 
    as described in the section [Creating a Sampler](#creating-a-sampler).
    
    __Example:__
    
    ~~~ .python
    src = pool['bark.ogg']
    my_sampler = Sampler(src, volume=.5)
    my_sampler.play()
    ~~~
    
    [TOC]
    
    ## Things to know
    
    ### Creating a Sampler
    
    You generally create a `Sampler` with the `Sampler()` factory function, which
    takes the full path to a sound file as the first argument.
    
    ~~~ .python
    src = pool['bark.ogg']
    my_sampler = Sampler(src)
    ~~~
    
    Optionally, you can pass [Playback keywords](#playback-keywords) to `Sampler()`
    to set the default behavior:
    
    ~~~ .python
    src = pool['bark.ogg']
    my_sampler = Sampler(src, volume=.5)
    ~~~
    
    ### Sampling rate
    
    If you find that your sample plays too slowly (low pitch) or too quickly (high
    pitch), make sure that the sampling rate of your sample matches the sampling
    rate of the sampler back-end as specified under backend settings.
    
    ### Supported file formats
    
    Sound files in `.wav`, `.mp3`, and `.ogg` format are supported. If you need to
    convert samples from a different format, you can use
    [Audacity](http://sourceforge.net/projects/audacity/).
    
    ### Playback keywords
    
    Functions that accept `**playback_args` take the following keyword arguments:
    
    - `volume` specifies a volume between `0.0` (silent) and `1.0` (maximum).
    - `pitch` specifies a pitch (or playback speed), where values > 1 indicate a
      higher pitch, and values < 1 indicate a lower pitch.
    - `pan` specifies a panning, where values < 0 indicate panning to the left, and
      values > 0 indicate panning to the right. Alternatively, you can set pan to
      'left' or 'right' to play only a single channel.
    - `duration` specifies the duration of the sound in milliseconds, or is set to
      `0` or `None` to play the full sound.
    - `fade_in` specifies the fade-in time (or attack) of the sound, or is set to
      `0` or `None` to disable fade-in.
    - `block` indicates whether the experiment should block (`True`) during
      playback or not (`False`).
    
    ~~~ .python
    src = pool['bark.ogg']
    my_sampler = Sampler(src)
    my_sampler.play(volume=.5, pan='left')
    ~~~
    
    Playback keywords only affect the current operation (except when passed to
    `Sampler()` when creating the object). To change the behavior for all
    subsequent operations, set the playback properties directly:
    
    ~~~ .python
    src = pool['bark.ogg']
    my_sampler = Sampler(src)
    my_sampler.volume = .5
    my_sampler.pan = 'left'
    my_sampler.play()
    ~~~
    
    Or pass the playback keywords to `Sampler()` when creating the object:
    
    ~~~ .python
    src = pool['bark.ogg']
    my_sampler = Sampler(src, volume=.5, pan='left')
    my_sampler.play()
    ~~~
    """
    def __init__(self, experiment, src, **playback_args):
        r"""Constructor to create a new SAMPLER object. You do not generally
        call this constructor directly, but use the `Sampler()` function,
        which
        is described here: [/python/sampler/]().

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        src : unicode, str
            The full path to a `.wav` or `.ogg` file.
        **playback_args : dict
            Optional [playback keywords](#playback-keywords) that will be used
            as the default for this SAMPLER object.


        Examples
        --------
        >>> src = pool['my_sound.ogg']
        >>> my_sampler = Sampler(src, volume=.5)
        """
        self.experiment = experiment
        Backend.__init__(self, configurables={
            'volume': self.assert_numeric,
            'pan': self.assert_pan,
            'pitch': self.assert_numeric,
            'duration': self.assert_numeric_or_None,
            'fade_in': self.assert_numeric_or_None,
            'block': self.assert_bool,
        }, **playback_args)

    def default_config(self):

        return {
            'volume': 1.,
            'pan': 0.,
            'pitch': 1.,
            'duration': None,
            'fade_in': None,
            'block': False,
        }

    def assert_pan(self, key, val):

        if not isinstance(val, int) and not isinstance(val, float) \
                and val not in ['left', 'right']:
            raise InvalidValue(
                'pan should be numeric, \'left\', or \'right\', not %s' % val)

    @configurable
    def play(self, **playback_args):
        r"""Plays the sound.

        Parameters
        ----------
        **playback_args : dict
            Optional [playback keywords](#playback-keywords) that will be used
            for this call to `Sampler.play()`. This does not affect subsequent
            operations.


        Examples
        --------
        >>> src = pool['my_sound.ogg']
        >>> my_sampler = Sampler(src)
        >>> my_sampler.play(pitch=.5, block=True)
        """
        raise NotImplementedError()

    def stop(self):
        r"""Stops the currently playing sound (if any).

        Examples
        --------
        >>> src = pool['my_sound.ogg']
        >>> my_sampler = Sampler(src)
        >>> my_sampler.play()
        >>> sleep(100)
        >>> my_sampler.stop()
        """
        raise NotImplementedError()

    def pause(self):
        r"""Pauses playback (if any).

        Examples
        --------
        >>> src = pool['my_sound.ogg']
        >>> my_sampler = Sampler(src)
        >>> my_sampler.play()
        >>> sleep(100)
        >>> my_sampler.pause()
        >>> sleep(100)
        >>> my_sampler.resume()
        """
        raise NotImplementedError()

    def resume(self):
        r"""Resumes playback (if any).

        Examples
        --------
        >>> src = pool['my_sound.ogg']
        >>> my_sampler = Sampler(src)
        >>> my_sampler.play()
        >>> sleep(100)
        >>> my_sampler.pause()
        >>> sleep(100)
        >>> my_sampler.resume()
        """
        raise NotImplementedError()

    def is_playing(self):
        r"""Checks if a sound is currently playing.

        Returns
        -------
        bool
            True if a sound is playing, False if not.

        Examples
        --------
        >>> src = pool['my_sound.ogg']
        >>> my_sampler = Sampler(src)
        >>> my_sampler.play()
        >>> sleep(100)
        >>> if my_sampler.is_playing():
        >>>         print('The sampler is still playing!')
        """
        raise NotImplementedError()

    def wait(self):
        r"""Blocks until the sound has finished playing or returns right away
        if no sound is playing.

        Examples
        --------
        >>> src = pool['my_sound.ogg']
        >>> my_sampler = Sampler(src)
        >>> my_sampler.play()
        >>> my_sampler.wait()
        >>> print('The sampler is finished!')
        """
        raise NotImplementedError()

    @staticmethod
    def init_sound(experiment):
        r"""Initializes the pygame mixer before the experiment begins.

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        """
        raise NotImplementedError()

    @staticmethod
    def close_sound(experiment):
        r"""Closes the mixer after the experiment is finished.

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        """
        raise NotImplementedError()


# Non PEP-8 alias for backwards compatibility
sampler = Sampler



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.keyboard_response import KeyboardResponse as \
    KeyboardResponseRuntime
from libqtopensesame.validators import TimeoutValidator
from libqtopensesame.items.qtplugin import QtPlugin
from openexp.keyboard import Keyboard
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.translate import translation_context
_ = translation_context('keyboard_response', category='item')


class KeyboardResponse(KeyboardResponseRuntime, QtPlugin):
    """keyboard_response item GUI"""
    
    description = _('Collects keyboard responses')
    help_url = 'manual/response/keyboard'
    lazy_init = True

    def __init__(self, name, experiment, string=None):
        KeyboardResponseRuntime.__init__(self, name, experiment, string)
        QtPlugin.__init__(self)

    def init_edit_widget(self):
        super().init_edit_widget(stretch=True)
        # Use auto-controls for most stuff
        self.add_line_edit_control(
            var='correct_response',
            label=_('Correct response'),
            info=_('Leave empty to use "correct_response"'))
        self.add_line_edit_control(
            var='allowed_responses',
            label=_('Allowed responses'),
            info=_('Separated by semicolons, e.g. "z;/"'))
        self.add_line_edit_control(
            var='timeout',
            label=_('Timeout'),
            info=_('In milliseconds or "infinite"'),
            validator=TimeoutValidator(self))
        self.add_combobox_control(
            var='event_type',
            label=_('Event type'),
            options=['keypress', 'keyrelease'])
        self.add_checkbox_control('flush', _('Flush pending key events'))
        # List available keys
        button_list_keys = QtWidgets.QPushButton(
            self.theme.qicon('help-about'),
            _('List available keys'))
        button_list_keys.setIconSize(QtCore.QSize(16, 16))
        button_list_keys.clicked.connect(self.list_keys)
        self.add_control('', button_list_keys)

    def list_keys(self):
        """Show a dialog with available key names"""
        my_keyboard = Keyboard(self.experiment)
        keylist = filter(lambda key: key.strip(), my_keyboard.valid_keys())
        md = '# ' + _('Key names') + '\n\n' \
            + _('The following key names are valid:') + '\n\n- ' \
            + '\n- '.join(['`%s`' % key for key in keylist])
        self.tabwidget.open_markdown(md, icon='os-keyboard_response',
                                     title=_('Key names'))


# Alias for backwards compatibility
keyboard_response = KeyboardResponse



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import pygame
from collections import OrderedDict
from openexp._canvas.canvas import Canvas
from openexp.backend import configurable
from expyriment import io, control, stimuli
from libopensesame.oslogging import oslogger
from openexp._coordinates.xpyriment import Xpyriment as XpyrimentCoordinates


class Xpyriment(Canvas, XpyrimentCoordinates):

    r"""This is a canvas backend built on top of Expyriment. For function
    specifications and docstrings, see `openexp._canvas.canvas`.
    """
    settings = {
        u"expyriment_opengl": {
            u"name": u"Use OpenGL",
            u"description": u"Use OpenGL mode for better temporal precision",
            u"default": u"yes"
        }
    }

    def __init__(self, experiment, auto_prepare=True, **style_args):

        Canvas.__init__(self, experiment, auto_prepare=auto_prepare,
                        **style_args)
        XpyrimentCoordinates.__init__(self)
        self.clear()

    @configurable
    def clear(self):

        self._elements = OrderedDict()
        self._set_background()

    def set_config(self, **cfg):

        Canvas.set_config(self, **cfg)
        if u'background_color' in cfg:
            self._set_background()

    def show(self):

        # First create a flat list of all elements, including elements that are
        # part of a group. Then we show them, clearing the display on the first,
        # and updating the display on the last.
        t0 = self.experiment.clock.time()
        elements = [e for g in self._elements.values() if g.visible for e in g]
        self._background.present(clear=True, update=not elements)
        while elements:
            e = elements.pop(0)
            e.show(clear=False, update=not elements)
        t1 = self.experiment.clock.time()
        if t1 - t0 > self.MAX_SHOW_DT:
            oslogger.warning('Canvas.show() took {0} ms'.format(t1 - t0))
        return t1

    def _set_background(self):

        self._background = stimuli.BlankScreen(
            colour=self.background_color.backend_color
        )
        self._background.preload()

    @staticmethod
    def init_display(experiment):

        # Configure Expyriment
        io.defaults.mouse_track_button_events = False
        control.defaults.initialize_delay = 0
        control.defaults.event_logging = 0
        control.defaults.window_mode = experiment.var.fullscreen != u'yes'
        control.defaults.fast_quit = True
        control.defaults.window_size = experiment.resolution()
        control.defaults.auto_create_subject_id = True
        control.defaults.open_gl = experiment.var.get(u'expyriment_opengl',
                                                      xpyriment.settings[u'expyriment_opengl'][u'default']) == u'yes'
        control.defaults.audiosystem_sample_rate = experiment.var.sound_freq
        control.defaults.audiosystem_bit_depth = \
            experiment.var.sound_sample_size
        control.defaults.audiosystem_channels = experiment.var.sound_channels
        control.defaults.audiosystem_buffer_size = \
            experiment.var.sound_buf_size
        control.defaults.audiosystem_autostart = False
        # Initialize. If Expyriment jumps into interactive mode, it reads from
        # the stdin, and crashes. Thus we explicitly disable the interactive-
        # mode detection. The is_interactive_mode() function is located
        # differently in different versions of expyriment, and hence we monkey
        # patch it twice.
        control._experiment_control.misc.is_interactive_mode = \
            control._experiment_control.is_interactive_mode = \
            lambda: False
        exp = control.initialize()
        experiment.window = exp.screen._surface
        experiment.expyriment = exp
        # TODO: In order to set the window title and to allow mouse responses we
        # need to bypass expyriment for now
        pygame.display.set_caption(u'OpenSesame (Expyriment backend)')
        pygame.event.set_allowed(pygame.MOUSEBUTTONDOWN)
        pygame.event.set_allowed(pygame.MOUSEBUTTONUP)

    @staticmethod
    def close_display(experiment):

        control.end()


# Non PEP-8 alias for backwards compatibility
xpyriment = Xpyriment



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame import metadata
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.base_component import BaseComponent
from libqtopensesame.misc import theme


class QtOpenSesameRun(QtWidgets.QMainWindow, BaseComponent):
    """Implements the GUI for opensesamerun.
    
    Parameters
    ----------
    options:
        Command-line arguments passed to opensesamerun, as parsed by
        `libopensesame.misc.opensesamerun_options()`.
    parent: optional
    """
    def __init__(self, options, parent=None):
        QtWidgets.QMainWindow.__init__(self, parent)
        # Setup the UI
        self.load_ui('misc.opensesamerun')
        self.ui.button_run.clicked.connect(self.run)
        self.ui.label_header.setText(
            self.ui.label_header.text() % metadata.__version__)
        self.theme = theme.theme(self)
        self.ui.button_browse_experiment.clicked.connect(
            self.browse_experiment)
        self.ui.button_browse_logfile.clicked.connect(self.browse_logfile)
        self.options = options
        # Fill the GUI controls based on the options
        self.ui.edit_experiment.setText(self.options.experiment)
        self.ui.checkbox_fullscreen.setChecked(self.options.fullscreen)
        self.ui.spinbox_subject_nr.setValue(int(self.options.subject))
        self.ui.edit_logfile.setText(self.options.logfile)

    def browse_experiment(self):
        """Locates the experiment file."""
        file_type_filter = \
            "OpenSesame files (*.osexp *.opensesame.tar.gz *.opensesame);;OpenSesame script and file pool (*.opensesame.tar.gz);;OpenSesame script (*.opensesame)"
        path = QtWidgets.QFileDialog.getOpenFileName(
            self, "Open experiment file", filter=file_type_filter)
        # In PyQt5, the QFileDialog.getOpenFileName returns a tuple instead of
        # a string, of which the first position contains the path.
        if isinstance(path, tuple):
            path = path[0]
        if path == "":
            return
        self.ui.edit_experiment.setText(path)

    def browse_logfile(self):
        """Locates the logfile."""
        path = QtWidgets.QFileDialog.getSaveFileName(
            self, "Choose a location for the logfile")
        # In PyQt5, the QFileDialog.getOpenFileName returns a tuple instead of
        # a string, of which the first position contains the path.
        if isinstance(path, tuple):
            path = path[0]
        if path == "":
            return
        self.ui.edit_logfile.setText(path)

    def show(self):
        """Sets the run flag to false."""
        self.run = False
        QtWidgets.QMainWindow.show(self)

    def run(self):
        """Does not actual run the experiment, but marks the application for
        running later.
        """
        self.run = True
        self.options.experiment = str(self.ui.edit_experiment.text())
        self.options.subject = self.ui.spinbox_subject_nr.value()
        self.options.logfile = str(self.ui.edit_logfile.text())
        self.options.fullscreen = self.ui.checkbox_fullscreen.isChecked()
        self.options.custom_resolution = \
            self.ui.checkbox_custom_resolution.isChecked()
        self.options.width = self.ui.spinbox_width.value()
        self.options.height = self.ui.spinbox_height.value()
        self.close()


# Alias for backwards compatibility
qtopensesamerun = QtOpenSesameRun


if __name__ == '__main__':
    from libqtopensesame import __main__
    __main__.opensesamerun()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import subprocess
import tempfile
from libopensesame.exceptions import OSException
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from libqtopensesame.runners import BaseRunner
from qtpy import QtWidgets
import time


class ExternalRunner(BaseRunner):

    """Runs an experiment using opensesamerun."""
    def execute(self):
        """See base_runner.execute()."""
        # Temporary file for the standard output and experiment
        self.stdout = tempfile.mktemp(suffix=".stdout")
        if self.experiment.experiment_path is None:
            return OSException(
                'Please save your experiment first, before running it using '
                'opensesamerun')
        self.path = os.path.join(self.experiment.experiment_path,
                                 '.opensesamerun-tmp.osexp')
        self.experiment.save(self.path, True)
        oslogger.debug("experiment saved as '%s'" % self.path)
        # Determine the name of the executable. The executable depends on the
        # platform, package, and Python version.\
        if cfg.opensesamerun_exec == '':
            # Is there a direct executable?
            if os.path.exists('opensesamerun.exe'):
                self.cmd = ['opensesamerun.exe']
            elif os.path.exists('opensesamerun.bat'):
                self.cmd = ['opensesamerun.bat']
            elif os.path.exists('opensesamerun'):
                self.cmd = ['opensesamerun']
            # Or is there a Python interpreter and a script with a known name?
            elif os.path.exists('python.exe') and \
                    os.path.exists(os.path.join('Scripts', 'opensesamerun')):
                self.cmd = ['python.exe',
                            os.path.join('Scripts', 'opensesamerun')]
            elif os.path.exists('python.exe') and \
                    os.path.exists(os.path.join('Scripts',
                                                'opensesamerun-script.py')):
                self.cmd = ['python.exe',
                            os.path.join('Scripts', 'opensesamerun-script.py')]
            else:
                return OSException(
                    'Failed to locate opensesamerun. Try selecting a '
                    'different runner under Preferences.')
        else:
            self.cmd = cfg.opensesamerun_exec.split()
        self.cmd += [self.path, 
                     "--logfile=%s" % self.experiment.logfile,
                     "--subject=%s" % self.experiment.var.subject_nr]
        if oslogger.debug_mode:
            self.cmd.append("--debug")
        if self.experiment.var.fullscreen == 'yes':
            self.cmd.append("--fullscreen")
        oslogger.debug("spawning opensesamerun as a separate process")
        # Call opensesamerun and wait for the process to complete
        try:
            p = subprocess.Popen(self.cmd, stdout=open(self.stdout, "w"))
        except Exception as e:
            try:
                os.remove(self.path)
                os.remove(self.stdout)
            except:
                pass
            return e
        # Wait for OpenSesame run to complete, process events in the meantime,
        # to make sure that the new process is shown (otherwise it will crash
        # on Windows).
        retcode = None
        while retcode is None:
            retcode = p.poll()
            QtWidgets.QApplication.processEvents()
            time.sleep(1)
        oslogger.debug("opensesamerun returned %d" % retcode)
        print()
        print(open(self.stdout, "r").read())
        print()
        # Clean up the temporary file
        try:
            os.remove(self.path)
            os.remove(self.stdout)
        except:
            pass
        return None

    def workspace_globals(self):

        return {'logfile': self.experiment.logfile}
